# ------------------------------------------------------------------------
#
# gnuplot.tcl --
#
# Configuration file for PWTK's Gnuplotting.
#
# This configuration file is sourced from the ::pwtk::gp namespace.
#
# ------------------------------------------------------------------------


#****iv*
# ARRAY
#   ::pwtk::gp::options
#
# PURPOSE
#   PWTK's pre-set options for gnuplot plots (as to make plots nicer).
#
#******

array set options {
    samples    200
    mxtics     2
    mytics     2    
    grid       {lt -1 lc '#bbbbbb'}
    style.fill {transparent solid 0.4}
    xzeroaxis  {lt -1}
    yzeroaxis  {lt -1}
}


#****iv*
# ARRAY
#   ::pwtk::gp::mapOptions
#
# PURPOSE
#   PWTK's pre-set options for gnuplot heatmaps.
#
#******

array set mapOptions {
    samples    200
    isosamples 50,50
    pm3d       {map interpolate 0,0}
    
    contour    base
    cntrparam  {levels auto 10}
    
    key        off
    
    mxtics     2
    mytics     2    
}
#    xzeroaxis  {lt -1}
#    yzeroaxis  {lt -1}

#****iv*
# ARRAY
#   ::pwtk::gp::surfOptions
#
# PURPOSE
#   PWTK's pre-set options for gnuplot 3D surface plots.
#
#******

# consider:
#    for        {[i=1:20] linetype i linecolor black}

array set surfOptions {
    samples    200
    isosamples 50,50
    hidden3d   defaults
    pm3d       {at s interpolate 0,0 lighting}

    contour    surface
    cntrparam  {levels auto 10}

    key        off
    
    mxtics     2
    mytics     2    
}
#    xzeroaxis  {lt -1}
#    yzeroaxis  {lt -1}

#****iv*
# ARRAY
#   ::pwtk::gp::surfmapOptions
#
# PURPOSE
#   PWTK's pre-set options for gnuplot 3D surface with map@base plots.
#
#******

array set surfmapOptions [array get surfOptions]
array set surfmapOptions {
    pm3d       {at bs interpolate 0,0 lighting}
    contour    both
}


#****iv*
# ARRAY
#   ::pwtk::gp::palette
#
# PURPOSE
#   12 PWTK's pre-defined pm3d palettes.
#
#******


array set ::pwtk::gp::palette {
    gnuplot {rgbformulae 7,5,15}
    
    green-red-violet {rgbformulae 3,11,6}
    
    green-blue-red-orange {defined (0 "green", 1 "blue", 2 "red", 3 "orange")}
    
    rgb {defined (0  1 0 0,  1  0 1 0,  2  0 0 1)}
    
    pale-rainbow {rgbformulae 33,13,10}
    
    rainbow {file "-"
        0.750   0.243   0.95
        0.000   0.322   0.95
        0.173   0.816   0.95
        0.408   0.950   0.176
        0.95    0.950   0.000
        1.000   0.635   0.0
        1.000   0.0     0.0
        end}

    blue-white-red {file "-"
        0.0 0.0 1.0
        0.63 0.63 1.0
        1.0 1.0 1.0
        1.0 0.63 0.63
        1.0 0.0 0.0
        end}

    dark-blue-white-red {file "-"
        0.0 0.0 0.5
        0.6 0.6 1.0
        1.0 1.0 1.0
        1.0 0.6 0.6
        0.5 0.0 0.0
        end}
    
    ocean {rgbformulae 23,28,3}
    
    AFM {rgbformulae 34,35,36 }
    
    STM {file "-"
       0.0 0.0 0.0
       0.48 0.14 0.02
       0.78 0.63 0.07
       1.0 1.0 1.0
       end}

    grayscale {defined (0  0 0 0, 1  1 1 1)}
}


#****iv*
# LIST
#   ::pwtk::gp::known_options
#
# DESCRIPTION
#   The purpose of "known_options" is to have a "logically" sorted
#   options in PWTK's generated Gnuplot scripts (Gnuplot's options are
#   options that are set with the Gnuplot's "set" command).
#
# BEWARE
#   Options consisting of more than one word, such as "format x" is
#   specified as format.x
#
#******
set known_options {
    tmargin     
    rmargin 
    bmargin 
    lmargin 

    samples

    isosamples    
    hidden3d
    pm3d
    palette
    contour
    cntrparam
    cbtics
    cbrange

    origin
    view

    cbtics
    xtics 
    ytics
    ztics
    mxtics 
    mytics
    mztics
    x2tics 
    y2tics
    mx2tics 
    my2tics

    cbrange
    xrange 
    yrange
    zrange
    x2range     
    y2range 
    
    format.x 
    format.y
    format.z
    format.x2 
    format.y2 
    format.cb
    
    xzeroaxis 
    yzeroaxis 

    grid 
    style.fill

    xlabel  
    ylabel
    zlabel
    x2label 
    y2label 

    key     
    title       
}

#****iv*
# ARRAY
#   ::pwtk::gp::pwtk2gp
#
# PURPOSE
#   Array holding the matching between PWTK and Gnuplot terminal
#   names. Its structure is the following:
#
#      pwtk2gp(pwtkName) == gnuplotName
#
# BEWARE
#   The empty Gnuplot's terminal, {}, stands for a default Gnuplot's terminal
#   (which is probably a window terminal); it's PWTK name is "win".
#******

array set pwtk2gp {
    ps  postscript
    eps postscript
    png pngcairo
    pdf pdfcairo
    win {}
}


#****iv*
# ARRAY
#   ::pwtk::gp::gp
#
# PURPOSE
#   The "main" config data (such as terminals, fonts...) for Gnuplot plots.
#
# BEWARE
#   1. PWTK's terminal names eps, pdf, and png are named as
#      postscript, pdfcairo, and pngcairo in Gnuplot
#
#   2. during initialization: gp(terminals) is set to a merge of
#      file.terminals, win.terminals, txt.terminals
#
#   3. an empty win.terminal stands for Gnuplot's defult terminal
#      (which is most-likely a window-type terminal)
#******

array set gp {
    file.terminals {postscript pdfcairo svg pngcairo}
    win.terminals  {qt wxt x11 {}}
    txt.terminals  {dumb}
    
    font         Arial
    fontSize     14
    svg.fontSize 18
    png.fontSize 15
    
    eps.portrait.size   0.8,0.4
    eps.landscape.size  0.53,0.6
    pdf.size   4,3
    png.size   800,575
    #size       640,460
    size       800,575
    
    lw         1.1
}
#pause      mouse
#pauseText  "Click"

# defaults for "with" for datafiles & functions
array set gp {
    datafile_with {w linespoints lw 3 pt 6}
    function_with {w l lw 3}
}


#****iv*
# CONTINUED
#   ::pwtk::gp::gp
#
# Config for multiplot:
#
#   npmax == maximum number of plots per page
#   np    == number of plots per pag
#   npN   == limiting np values for determinning nxN values, if np > npN ==> nx = nxN
#   nxN   == number of plots in X-dir for a given npN
#   font.nxN == font size for a given nxN
#
#   spin.*** == spin-PDOS-like plots (plots with positive and negative Y-values,
#               i.e., f(x) & -f(x))
#******

array set gp {
    eps.multisize  1.0,1.0
    pdf.multisize  11.69,8.27
    multisize      1200,800

    landscape.npmax 25
    portrait.npmax  28

    landscape.np1 4
    landscape.np2 9
    landscape.np3 20
    landscape.np4 30

    landscape.nx0 2
    landscape.nx1 3
    landscape.nx2 4
    landscape.nx3 5
    landscape.nx4 6

    portrait.np1 4
    portrait.np2 10
    portrait.np3 21
    portrait.np4 32
    
    portrait.nx0 1
    portrait.nx1 2
    portrait.nx2 3
    portrait.nx3 4
    portrait.nx4 5

    spin.landscape.np1 4
    spin.landscape.np2 9
    spin.landscape.np3 20
    spin.landscape.np4 30

    spin.landscape.nx0 2
    spin.landscape.nx1 3
    spin.landscape.nx2 4
    spin.landscape.nx3 5
    spin.landscape.nx4 6
    
    spin.portrait.np1 4
    spin.portrait.np2 9
    spin.portrait.np3 20
    spin.portrait.np4 30

    spin.portrait.nx0 1
    spin.portrait.nx1 2
    spin.portrait.nx2 3
    spin.portrait.nx3 4
    spin.portrait.nx4 5

    fontSize.nx0 14
    fontSize.nx1 11
    fontSize.nx2 9    
    fontSize.nx3 8
    fontSize.nx4 7
}
