#
# $RCSfile: parse_opt.tcl,v $ --
#
#      This file contains some procs for parsing options, used
#      internally by PWTK.
#
# Copyright (c) 2023  Anton Kokalj   Email: tone.kokalj@ijs.si
#
#
# This file is distributed under the terms of the GNU General Public
# License. See the file `LICENCE' in the root directory of the present
# distribution, or http://www.gnu.org/copyleft/gpl.txt .
#
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# ANTON KOKALJ BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
# AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
#
# $Id: parse_opt.tcl,v 1.3 2023/01/20 09:45:01 kokalj Exp $ 
#


#****if* Internals/::pwtk::parseOpt_
# SYNOPSIS
proc ::pwtk::parseOpt_ {} {    
    # DESCRIPTION    
    #   A convenience routine for parsing options for procs that are
    #   of "proc name {args} {...}" form, where the usage is as:
    #
    #     name -opt1 -opt2 val -opt3 -val   arg1  arg2 arg3
    #          ^^^^^^^^^^^^^^^^^^^^^^^^^^   ^^^^^^^^^^^^^^^
    #           |                                 |
    #           +--- options and their values     + extra agruments that follow after options
    # BEWARE
    # In addition to the args variable, the following variables must
    # also be defined to use this proc:    
    # * options ... option specs for ::cmdline::getoptions
    # * usage   ... usage string to print for "usage" error
    #
    # and either:
    # * narg    ... number of arguments that follow the options (in above example narg == 3);
    #               if narg < 0 then no checking is performed
    # or:
    # * nargmin, nargmax ... min and max number of allowed arguments that follow the options
    #                        if nargmax < 0 then no checking is performed
    #
    #
    # The parsed opions are stored in "opt" array.
    #
    # SOURCE
    
    uplevel 1 {
        foreach var {args options usage} {            
            if { ! [info exists $var] } {
                ::error "coding error: variable $var does not exist"
            }
        }
        set args_ $args 
        array set opt [::cmdline::getoptions args $options "error while executing:\n\n[pwtk::procName] $args_\n\n[::pwtk::procName] options:"]
        
        if { [info exists narg] } {
            if { $narg >= 0 && [llength $args] != $narg } {
                # nargs
                ::pwtk::error "wrong # args, should be:\n\n[::pwtk::procName] $usage\n\nbut got: [::pwtk::procName] $args_" 1
            }
        } elseif { [info exists nargmin] && [info exists nargmax] } {
            # nargmin nargmax
            if { [llength $args] < $nargmin } {
                ::pwtk::error "wrong # args, should be:\n\n[::pwtk::procName] $usage\n\nbut got: [::pwtk::procName] $args_" 1
            }
            if { $nargmax >= 0 && [llength $args] > $nargmax } {
                ::pwtk::error "wrong # args, should be: \n\n[::pwtk::procName] $usage\n\nbut got: [::pwtk::procName] $args_" 1
            }
            # if nargmax < 0 ==> no checking
        } else {
            ::error "coding error: neither narg nor nargmin & nargmax exist"
        }        
    }
}
#******


#****if* Internals/::pwtk::parseKnownOpt_
# SYNOPSIS
proc ::pwtk::parseKnownOpt_ {} {    
    # DESCRIPTION
    #   Similar to ::pwtk::parseOpt_ but ignores any unknown options,
    #   i.e., the difference between ::pwtk::parseKnownOpt_ and ::pwtk::parseOpt_
    #   is that the former uses ::cmdline::getKnownOptions and the latter
    #   ::cmdline::getoptions.
    #
    #   A convenience routine for parsing known options for procs that are
    #   of "proc name {args} {...}" form, where the usage is as:
    #
    #     name -opt1 -opt2 val -opt3 -val   arg1  arg2 arg3
    #          ^^^^^^^^^^^^^^^^^^^^^^^^^^   ^^^^^^^^^^^^^^^
    #           |                                 |
    #           +--- options and their values     + extra agruments that follow after options
    # BEWARE
    # In addition to the args variable, the following variables must
    # also be defined to use this proc:    
    # * options ... option specs for ::cmdline::getoptions
    # * usage   ... usage string to print for "usage" error
    #
    # and either:
    # * narg    ... number of arguments that follow the options (in above example narg == 3);
    #               if narg < 0 then no checking is performed
    # or:
    # * nargmin, nargmax ... min and max number of allowed arguments that follow the options
    #                        if nargmax < 0 then no checking is performed
    #    
    # The parsed opions are stored in "opt" array.
    #
    # SOURCE
    
    uplevel 1 {
        foreach var {args options usage} {            
            if { ! [info exists $var] } {
                ::error "coding error: variable $var does not exist"
            }
        }
        set args_ $args 
        array set opt [::cmdline::getKnownOptions args $options "error while executing:\n\n[pwtk::procName] $args_\n\n[::pwtk::procName] options:"]
        
        if { [info exists narg] } {
            if { $narg >= 0 && [llength $args] != $narg } {
                # nargs
                ::pwtk::error "wrong # args, should be:\n\n[::pwtk::procName] $usage\n\nbut got: [::pwtk::procName] $args_" 1
            }
        } elseif { [info exists nargmin] && [info exists nargmax] } {
            # nargmin nargmax
            if { [llength $args] < $nargmin } {
                ::pwtk::error "wrong # args, should be:\n\n[::pwtk::procName] $usage\n\nbut got: [::pwtk::procName] $args_" 1
            }
            if { $nargmax >= 0 && [llength $args] > $nargmax } {
                ::pwtk::error "wrong # args, should be: \n\n[::pwtk::procName] $usage\n\nbut got: [::pwtk::procName] $args_" 1
            }
            # if nargmax < 0 ==> no checking
        } else {
            ::error "coding error: neither narg nor nargmin & nargmax exist"
        }        
    }
}
#******


#****if* Internals/::pwtk::checkOType_
# SYNOPSIS
proc ::pwtk::checkOType_ {option value type {typeTxt {}}} {
    # DESCRIPTION
    #   Check that the value of an option is of correct type.
    #
    #   Note that empty string matches the type (for the opposite
    #   behaviour use the strict version, checkOTypeStrict_)
    #
    # ARGUMENTS
    # * option == name of the option
    # * value  == value of the option
    # * type   == type for the value
    # * typeTxt == [optional, default = $type] a textual-type of value that is printed
    #              to error message (for example: type = double, typeTxt = number)
    # SOURCE
    if { $typeTxt == {} } {
        set typeTxt $type
    }
    set option -[string trim $option -]
    switch -exact -- [lindex $type 0] {
        number -
        numberlist {
            if { ! [eval ::pwtk::type::$type [list $value]] } {
                uplevel 1 ::pwtk::error [list "expected a $typeTxt for the value of the $option option, but got: \"$value\""] 1
            }
        }
        optionlist {
            set allowedValues [lrange $type 1 end]
            if { ! [::pwtk::type::optionlist $allowedValues $value] } {
                uplevel 1 ::pwtk::error [list "expected one of [join $allowedValues {, }] for the value of the $option option, but got: \"$value\""] 1
            }
        }
        default {
            if { ! [string is $type $value] } {
                uplevel 1 ::pwtk::error [list "expected a $typeTxt for the value of the $option option, but got: \"$value\""] 1
            }
        }
    }
}
#******


#****if* Internals/::pwtk::checkOTypeStrict_
# SYNOPSIS
proc ::pwtk::checkOTypeStrict_ {option value type {typeTxt {}}} {
    # DESCRIPTION
    #   A strict version of checkOType_, where empty string do not match the type.
    #
    # type can be: intList, doubleList, type accepted by "string is...."
    #
    # SOURCE
    if { $typeTxt == {} } {
        set typeTxt $type
    }
    set option -[string trim $option -]

    switch -exact -- [lindex $type 0] {
        number -
        numberlist {
            if { ! [eval ::pwtk::type::$type -strict [list $value]] } {
                uplevel 1 ::pwtk::error [list "expected a $typeTxt for the value of the $option option, but got: \"$value\""] 1
            }
        }
        optionlist {
            set allowedValues [lrange $type 1 end]
            if { ! [::pwtk::type::optionlist $allowedValues -strict $value] } {
                uplevel 1 ::pwtk::error [list "expected one of [join $allowedValues {, }] for the value of the $option option, but got: \"$value\""] 1
            }
        }
        default {
            if { ! [string is $type -strict $value] } {
                uplevel 1 ::pwtk::error [list "expected a $typeTxt for the value of the $option option, but got: \"$value\""] 1
            }
        }
    }
}
#******


#****if* Internals/::pwtk::trimImplicitOpts_
proc ::pwtk::trimImplicitOpts_ {optionListArgs} {
    # DESCRIPTION
    #   Trim implicit -?, -help, and -- options from the option-list
    #
    return [regsub -all {(\-\?|\-help|\-\-)} $optionListArgs {}]
}
#******
