#
# $RCSfile: scanpar.tcl,v $ --
#
#      This file contains utility commands for performing scan
#      calculations and tabulating & plotting the corresponding
#      results, in particular, commands: scanpar, write, finish, plot,
#      multiplot, heatmap, and splot.
#
# Copyright (c) 2023-2024  Anton Kokalj   Email: tone.kokalj@ijs.si
#
#
# This file is distributed under the terms of the GNU General Public
# License. See the file `LICENCE' in the root directory of the present
# distribution, or http://www.gnu.org/copyleft/gpl.txt .
#
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# ANTON KOKALJ BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
# AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
#
# $Id: scanpar.tcl,v 1.3 2023/02/03 17:00:01 kokalj Exp $ 
#


#****f* 5.8. Parameter Scans/::pwtk::scanpar
# SYNOPSIS
proc ::pwtk::scanpar {paramName values script} {
    # USAGE    
    #   ::pwtk::scanpar paramName values script
    #
    # ARGUMENTS
    # * paramName --- name of the parameter to scan
    # * values  --- list of parameter values to scan
    # * script  --- script to execute for each value of the parameter
    #
    # DESCRIPTION    
    #   The 'scanpar' command is analogous to a one-parameter Tcl
    #   "foreach" command and has a similar syntax, but (1) it keeps
    #   track of the scanned parameter to make the creation of
    #   datafiles a bit easier, and (2) uses input data stacking
    #   (::pwtk::input::push_pop) to restore the input data after the
    #   completion of the scan.
    #
    #   Below are two examples using the "foreach" and "scanpar"
    #   commands that are analogous.
    #
    # EXAMPLES
    #   # to illustrate what "scanpar" does, we first do an example
    #   # using the "foreach" Tcl command, and then repeat the same
    #   # example using "scanpar"
    #
    #   # 1. using "foreach"
    #
    #   write ecut.dat "# columns are: ecut data"
    #   input_pushpop {
    #      foreach ecut {10 14 18 22} {
    #         SYSTEM "ecutwfc = $ecut"
    #         runPW scf.ecut$ecut
    #         write ecut.dat "$k [pwo_totene scf.ecut$ecut.out]"
    #      }
    #   }
    #   plot -xl "ecutwfc (Ry)" -yl "Total energy (Ry)" ecut.dat
    #
    #   # 2. the same example using "scanpar"
    #
    #   scanpar ecut {10 14 18 22} {
    #      SYSTEM "ecutwfc = $ecut"
    #      runPW scf.ecut$ecut
    #      write ecut.dat [pwo_totene scf.ecut$ecut.out]
    #   }
    #   plot -xl "ecutwfc (Ry)" -yl "Total energy (Ry)" ecut.dat
    #
    # EXPLANATION
    #   In both examples above, identical "ecut.dat" datafile is
    #   created, but with "scanpar", the 1st-line comment is
    #   automatically created and the parameter $k is not specified in
    #   the "write" command because "write" adds $k automatically as
    #   it knows which parameter(s) is (are) being scanned with
    #   "scanpar".
    #
    # SOURCE

    variable scanpar
    variable scanparList
    
    upvar $paramName par

    if { [info exists scanparList] && ($paramName in $scanparList) } {
        ::pwtk::error  "scanpar already scans the \"$paramName\" parameter, name of the parameter must be unique" 1
    }

    ifnotempty scanparList {
        set d_ ""
        foreach p $scanparList {
            upvar $p p_
            append pt "${d_}$p = ${p_}"
            set d_ {, }
        }
        print "${pt}: scanning the paramater \"$paramName\" for values:   [join $values {, }]\n"
    } else {
        print "Scanning the paramater \"$paramName\" for values:   [join $values {, }]\n"
    }
    
    lappend scanparList $paramName
    input_pushpop {
        foreach par $values {
            # save the current value of the parameter (used by "write" if
            # called within the script)
            set scanpar($paramName) $par
            set code [catch {uplevel 1 $script} result]
            if {($code != 0) && ($code != 4)} { 
                break 
            }
        }
    }
    set scanparList [lrange $scanparList 0 end-1]
    if { $scanparList eq {} } {
        unset scanparList
        unset scanpar($paramName)
    }
    
    if { ($code == 0) || ($code == 3) || ($code == 4) } {
        return $result
    }
    return -code $code $result
}
#******


#****f* 5.8. Parameter Scans/::pwtk::write
# SYNOPSIS
proc ::pwtk::write {args} {
    # USAGE
    #   ::pwtk::write ?-a? DATAFILE LINE
    #
    # OPTION
    # * -a       --- in the first call of write on a given datafile, open the datafile in append mode (default = "w", i.e., truncate-and-write)
    #
    # ARGUMENTS
    # * DATAFILE --- name of the file for writing the data
    # * LINE     --- line-of-data to write to the datafile
    #
    # PURPOSE    
    #   "write" writes a line-of-data to a data file. If "write" is
    #   executed within the "scanpar" loops, the content of the line
    #   is prefixed with the list of scanning parameters. For example,
    #   the snipptet:
    #
    #        scanpar k {4 8} {
    #           scanpar ecut {20 30} {
    #              write energy.dat [pwo_totene scf.k$k.ecut$ecut.out]
    #           }
    #        }
    #
    #   writes the lines in the form:  $k  $ecut  energy
    # 
    # SOURCE
    
    variable write_fid
    variable scanpar
    variable scanparList
    
    # parse args

    set narg 2
    set usage "?-a? datafile line"
    set options {
        {a "open the datafile in the append mode"}
    }
    parseOpt_
    set mode     [expr { $opt(a) ? "a" : "w"}]
    set datafile [lindex $args 0]
    set line     [lindex $args 1]

    # write to datafile
    
    ifnotexist write_fid($datafile) {
        set write_fid($datafile) [open $datafile $mode]
        if  { [info exists scanparList] } {
            # write the first line as a comment documenting what columns are"
            puts $write_fid($datafile) "# columns are: $scanparList data"
        }
    }
    if { [info exists scanparList] && [string trim $line] ne {} } {
        # do not write "$params {}" line
        foreach pn $scanparList {
            upvar $pn p
            ifexist p {
                append params "$p "
            } else {                
                ifexist scanpar($pn) {
                    append params "$scanpar($pn) "
                }
            }
        }
        set line [concat $params $line]
    }    
    puts $write_fid($datafile) $line
    flush $write_fid($datafile)
    return $datafile
}
#******


#****f* 5.8. Parameter Scans/::pwtk::finish
# SYNOPSIS
proc ::pwtk::finish {datafiles} {
    # USAGE
    #   ::pwtk::finish DATAFILE1 ?DATAFILE2? ...
    #
    # ARGUMENTS
    # * DATAFILE1 --- 1st datafile to close
    # * DATAFILE2 --- 2nd datafile to close
    # * ...
    #
    # PURPOSE    
    #   "finish" closes the channels associated with the supplied
    #   datafiles that were built with the "write" command. Its
    #   function is analogous to the Tcl "close" comand.
    #
    #   If a datafile is not associated with an open channel, nothing
    #   is done and no error is returned.
    #
    # SOURCE
    variable write_fid

    set closed 0
    foreach df $datafiles {
        ifexist write_fid($df) {
            flush $write_fid($df)
            close $write_fid($df)
            unset write_fid($df)
            print "Data were written to datafile:   $df"
            set closed 1
        }
    }
    if { $closed } { puts "" }
}
#******


#***internal...
proc ::pwtk::plot_parseOpt_ {} {
    uplevel 1 {
        set nargmin 1
        set nargmax -1        
        set usage "?-xl XLABEL?  ?-yl YLABEL?  ?-xr XRANGE?  ?-yr YRANGE?  ?-xf XFORMAT?  ?-yf YFORMAT?  ?-u USING?  ?-w WITH?  ?-t TERMINAL?  ?-v?  datafile1 ?datafile2? ..."
        set options {
            {xl.arg {}  "X-axis label"}
            {yl.arg {}  "Y-axis label"}
            {xr.arg {}  "X-range"}
            {yr.arg {}  "Y-range"}
            {xf.arg {}  "format for X-axis (e.g. %.2f)"}
            {yf.arg {}  "format for Y-axis (e.g. %.2f)"}
            {u.arg  {}  "Gnuplot's \"using\" plot specification"}
            {w.arg  {} "Gnuplot's \"with\" plot specification"}
            {e.arg  {}  "extra Gnuplot commands/configuration"}
            {t.arg  qt  "terminal for the plot"}
            {v          "visualize generated plot image (for file terminals)"}
        }
        parseOpt_
    }
}


#****f* 5.8. Parameter Scans/::pwtk::plot
# SYNOPSIS
proc ::pwtk::plot {args} {    
    # USAGE
    #    ::pwtk::plot ?OPTIONS?  DATAFN1  ?DATAFN2? ...
    #
    # where DATAFN1, DATAFN2... are either datafiles or functions to plot,
    # and OPTIONS are:
    #
    #    -xl XLABEL
    #    -yl YLABEL
    #    -xf XFORMAT
    #    -yf YFORMAT
    #    -xr XRANGE
    #    -yr YRANGE
    #    -u  USING
    #    -w  WITH
    #    -e  EXTRA
    #    -t  TERMINAL
    #    -v
    #
    # OPTIONS
    # * -xl XLABEL  --- X-axis label
    # * -yl YLABEL  --- Y-axis label
    # * -xr XRANGE  --- X-axis range specified as XMIN:XMAX or {XMIN:XMAX} or "XMIN:XMAX"
    # * -yr YRANGE  --- Y-axis range specified as YMIN:YMAX or {YMIN:YMAX} or "YMIN:YMAX"
    # * -xf XFORMAT --- format for X-axis (e.g. %.2f)
    # * -yf YFORMAT --- format for Y-axis (e.g. %.2f)
    # * -u USING    --- Gnuplot's "using" plot specification
    # * -w WITH     --- Gnuplot's "with" plot specification
    # * -e EXTRA    --- extra Gnuplot commands/configuration, e.g., -e {set nokey; set title 'simple plot'}
    # * -t TERMINAL --- terminal for the plot
    # * -v          --- visualize the generated image(s) (for file terminals)    
    #
    # ARGUMENTS
    # * DATAFN1 --- the 1st datafile or function to plot
    # * DATAFN2 --- the 2nd datafile or function to plot
    # * ...
    #
    # PURPOSE
    #   "plot" plots data from one or more datafiles or functions on
    #   the same plot. If the datafiles were created with the
    #   ::pwtk::write command, they are closed prior to being plotted.
    #
    #   This is meant as a fast way of plotting simple
    #   data. For a more elaborate and configurable plotting, use the
    #   ::pwtk::gp::plot object-oriented interface instead.
    #
    #   For an example, see the example in ::pwtk::scanpar.
    #
    # RETURN VALUE
    # * the name of created image (for file terminals)
    # * an empty string (for non-file terminals)
    #
    # SOURCE

    variable write_fid
    variable plot_id
    
    # parse args

    plot_parseOpt_

    ifnotempty opt(u) { set opt(u) "u $opt(u)" }
    #ifnotempty opt(w) { set opt(w) "w $opt(w)" }
    ifset opt(xl) ''
    ifset opt(yl) ''
    
    # plot's name
    
    if { [llength $args] == 1 && [file exists $args] } {
        set head plot-[file rootname $args]
    } else {
        set id [incr ::pwtk::gp::plot_id]
        set head plot-$id
    }

    # plot
    
    finish $args
    
    foreach datafn $args {
        if { [file exists $datafn] } {
            # datafn is a file
            set width [expr { $opt(w) eq {} ? $::pwtk::gp::gp(datafile_with) : "w $opt(w)" }]
            append plot "[squote $datafn] $opt(u) $width, "
        } else {
            # datafn is a function, skip $opt(u)
            set width [expr { $opt(w) eq {} ? $::pwtk::gp::gp(function_with) : "w $opt(w)" }]
            append plot "$datafn $width, "
        }
    }
    set plot [string trim $plot {, }]

    set gp [::pwtk::gp::plot new $head.$opt(t)]
    $gp options [list xlabel [squote $opt(xl)] ylabel [squote $opt(yl)]]
    ifnotempty opt(xr) { $gp options [list xrange [rangequote $opt(xr)]] }
    ifnotempty opt(yr) { $gp options [list yrange [rangequote $opt(yr)]] }
    ifnotempty opt(xf) { $gp options [list format.x [squote $opt(xf)]] }
    ifnotempty opt(yf) { $gp options [list format.y [squote $opt(yf)]] }
    ifnotempty opt(e)  { $gp add $opt(e) }
    
    $gp plot $plot
    $gp exec
    set output [$gp outputs]
    $gp destroy
    
    displayPlots_ $opt(v) $opt(t) $output
    
    return $output
}
#******


#****f* 5.8. Parameter Scans/::pwtk::multiplot
# SYNOPSIS
proc ::pwtk::multiplot {args} {
    # USAGE
    #    ::pwtk::multiplot  ?OPTIONS?  DATAFN_LIST1  ?DATAFN_LIST2? ...
    #
    # where DATAFN_LIST1, DATAFN_LIST2... are a list of datafiles
    # and/or functions to plot, and OPTIONS are:
    #
    #    -xl XLABEL  
    #    -yl YLABEL  
    #    -xr XRANGE  
    #    -yr YRANGE  
    #    -xf XFORMAT 
    #    -yf YFORMAT 
    #    -u  USING    
    #    -w  WITH     
    #    -e  EXTRA    
    #    -t  TERMINAL 
    #    -v           
    #
    # OPTIONS
    # * -xl XLABEL  --- X-axis label
    # * -yl YLABEL  --- Y-axis label
    # * -xr XRANGE  --- X-axis range specified as XMIN:XMAX or {XMIN:XMAX} or "XMIN:XMAX"
    # * -yr YRANGE  --- Y-axis range specified as YMIN:YMAX or {YMIN:YMAX} or "YMIN:YMAX"
    # * -xf XFORMAT --- format for X-axis (e.g. %.2f)
    # * -yf YFORMAT --- format for Y-axis (e.g. %.2f)
    # * -u USING    --- Gnuplot's "using" plot specification
    # * -w WITH     --- Gnuplot's "with" plot specification
    # * -e EXTRA    --- extra Gnuplot commands/configuration, e.g., -e {set nokey; set title 'simple multiplot'}
    # * -t TERMINAL --- terminal for the plot
    # * -v          --- visualize the generated image(s) (for file terminals)    
    #
    # ARGUMENTS
    # * DATAFN_LIST1 --- list of datafiles and/or functions to plot in the 1st plot
    # * DATAFN_LIST2 --- list of datafiles and/or functions to plot in the 2nd plot
    # * ...
    #
    # PURPOSE
    #   "multiplot" plots data as the multiplot. Each list of data
    #   (DATAFN_LIST) is plotted in a separate plot, i.e.,
    #   DATAFN_LIST1 in plot-1, DATAFN_LIST2 in plot-2, etc.
    #
    #   If the datafiles were created with the ::pwtk::write command,
    #   they are closed prior to being plotted.
    #
    #   This is meant as a fast way of plotting multi-plots. For a
    #   more elaborate and configurable multi-plotting, use the
    #   ::pwtk::gp::multiplot object-oriented interface instead.
    #
    # EXAMPLE    
    #   scanpar k [seq 2 2 8] {
    #      scanpar ecut [seq 10 4 30] {
    #         set head scf.k$k.ecut$ecut
    #         SYSTEM "ecutwfc = $ecut"
    #         K_POINTS (automatic) "$k $k $k   1 1 1"
    #         runPW $head
    #         write k.$k.dat [pwo_totene $head.out]
    #      }
    #   }
    #   
    #   # plot each datafile separately and also all of them together
    #   
    #   multiplot -xl "ecutwfc (Ry)" -yl "Total energy (Ry)" -u 2:3 \
    #             k.2.dat k.4.dat k.6.dat k.8.dat { k.2.dat k.4.dat k.6.dat k.8.dat }
    #
    # RETURN VALUE
    # * a list of created images (for file terminals)
    # * an empty string (for non-file terminals)
    #
    # SOURCE

    variable write_fid
    variable multiplot_id
   
    # parse args

    plot_parseOpt_

    ifnotempty opt(u) { set opt(u) "u $opt(u)" }
    #ifnotempty opt(w) { set opt(w) "w $opt(w)" }
    ifset opt(xl) ''
    ifset opt(yl) ''

    # plot

    finish [concat {*}$args]
    set nplot 0

    foreach datafnList $args {
        foreach datafn $datafnList {
            if { [file exists $datafn] } {
                # datafn is a file
                set width [expr { $opt(w) eq {} ? $::pwtk::gp::gp(datafile_with) : "w $opt(w)" }]
                append plot($nplot) "[squote $datafn] $opt(u) $width, "
            } else {
                # datafn is a function, skip $opt(u)
                set width [expr { $opt(w) eq {} ? $::pwtk::gp::gp(function_with) : "w $opt(w)" }]
                append plot($nplot) "$datafn $width, "
            }
        }
        set plot($nplot) [string trim $plot($nplot) {, }]
        incr nplot
    }
    
    set id [incr ::pwtk::gp::multiplot_id]    
    set gp [::pwtk::gp::multiplot new -o landscape auto $nplot multiplot-$id.$opt(t)]
    
    $gp options [list xlabel [squote $opt(xl)] ylabel [squote $opt(yl)]]
    ifnotempty opt(xr) { $gp options [list xrange [rangequote $opt(xr)]] }
    ifnotempty opt(yr) { $gp options [list yrange [rangequote $opt(yr)]] }
    ifnotempty opt(xf) { $gp options [list format.x [squote $opt(xf)]] }
    ifnotempty opt(yf) { $gp options [list format.y [squote $opt(yf)]] }
    ifnotempty opt(e)  { $gp add $opt(e) }

    for {set i 0} {$i < $nplot} {incr i} {
        $gp plot $plot($i)
    }
    $gp exec
    set outputs [$gp outputs]
    $gp destroy

    displayPlots_ $opt(v) $opt(t) $outputs
    
    return $outputs
}
#******


# internal...
proc ::pwtk::splotParseOpt_ {} {
    uplevel 1 {
        set usage "?-pal PALETTE?  ?-contour?  ?-cc CONTOUR_COLOR?  ?-clw CONTOUR_LINEWIDTH?  ?-i INTERPOLATION?  ?-type TYPE? ?-xl XLABEL?  ?-yl YLABEL?  ?-zl ZLABEL?  ?-xr XRANGE?  ?-yr YRANGE?  ?-zr ZRANGE?  ?-cbr XBRANGE?  ?-xf XFORMAT?  ?-yf YFORMAT?  ?-zf YFORMAT?  ?-cbf CBFORMAT? ?-u USING?  ?-w WITH?  ?-t TERMINAL?  ?-v?  datafile"
        set options {
            {pal.arg  {} "color palette"}
            {contour     "plot contours"}
            {cc.arg   {} "color of contours"}
            {clw.arg  {} "contour linewidth"}
            {i.arg    {} "interpolation degree"}
            {type.arg surface "type of splot"}

            {xr.arg  {}  "X-range"}
            {yr.arg  {}  "Y-range"}
            {zr.arg  {}  "Z-range"}
            {cbr.arg {}  "CB-range"}
            
            {xl.arg  {}  "X-axis label"}
            {yl.arg  {}  "Y-axis label"}
            {zl.arg  {}  "Z-axis label"}
            {xf.arg  {}  "format for X-axis (e.g. %.2f)"}
            {yf.arg  {}  "format for Y-axis (e.g. %.2f)"}
            {zf.arg  {}  "format for Z-axis (e.g. %.2f)"}
            {cbf.arg {}  "format for CB-axis (e.g. %.2f)"}
            
            {u.arg   {}  "Gnuplot's \"using\" splot specification"}
            {w.arg   {}  "Gnuplot's \"with\" splot specification"}
            {e.arg   {}  "extra Gnuplot commands/configuration"}
            {t.arg   qt  "terminal for the plot"}
            {v           "visualize generated plot image (for file terminals)"}
        }
        parseOpt_

        # check for clw to be a nonnegreal number
        checkOType_ -clw $opt(clw) {number nonnegreal} "non-negative number"
        checkOType_ -i   $opt(i)   {numberlist integer} "list of integer numbers"
    }
}

# internal...
proc ::pwtk::splotOptionsAdd_ {} {
    uplevel 1 {
        foreach {option elem} {
            xlabel xl
            ylabel yl
            zlabel zl

            format.x xf
            format.y yf
            format.z zf
            format.cb cbf
        } {
            ifnotempty opt($elem) {
                $gp options [list $option [squote $opt($elem)]]
            }
        }
        foreach {option elem} {
            xrange xr
            yrange yr
            zrange zr
            cbrange cbr            
        } {
            ifnotempty opt($elem) {
                $gp options [list $option [rangequote $opt($elem)]]
            }
        }
        ifnotempty opt(i) { $gp options [list pm3d [list interpolate $opt(i)]] }
        ifnotempty opt(e) { $gp add $opt(e) }
    }
}


#****f* 5.8. Parameter Scans/::pwtk::heatmap
# SYNOPSIS
proc ::pwtk::heatmap {args} {    
    # USAGE
    #    ::pwtk::heatmap  ?OPTIONS?  DATAFN
    #
    # where DATAFN is either a datafile or a function to plot, and
    # OPTIONS are:
    #
    #    -pal PALETTE
    #    -i   INTERPOLATION
    #    -contour
    #    -cc  CONTOUR_COLOR
    #    -clw CONTOUR_LINEWIDTH
    #    -xl  XLABEL  
    #    -yl  YLABEL  
    #    -xr  XRANGE  
    #    -yr  YRANGE  
    #    -cbr CBRANGE  
    #    -xf  XFORMAT 
    #    -yf  YFORMAT 
    #    -cbf CBFORMAT 
    #    -u   USING    
    #    -w   WITH     
    #    -e   EXTRA    
    #    -t   TERMINAL 
    #    -v           
    #    
    # OPTIONS
    # Options specific for heatmaps:
    # * -pal PALETTE  --- color palette for the heatmap (PWTK has a dozen predefined palettes, see the image below)
    # * -i NX,NY      --- interpolation degree along X- and Y-axis
    #                     (default is set to auto-optimal interpolation)
    # * -contour      --- plot contours
    # * -cc  CONTOUR_COLOR     --- color of contours
    # * -clw CONTOUR_LINEWIDTH --- contour linewidth
    #
    # General options:
    # * -xl XLABEL    --- X-axis label
    # * -yl YLABEL    --- Y-axis label

    # * -xr XRANGE    --- X-axis range specified as XMIN:XMAX or {XMIN:XMAX} or "XMIN:XMAX"
    # * -yr YRANGE    --- Y-axis range specified as YMIN:YMAX or {YMIN:YMAX} or "YMIN:YMAX"
    # * -cbr CBRANGE  --- CB-axis range  specified as CBMIN:CBMAX or {CBMIN:CBMAX} or "CBMIN:CBMAX"
    
    # * -xf XFORMAT   --- format for X-axis (e.g. %.2f)
    # * -yf YFORMAT   --- format for Y-axis (e.g. %.2f)
    # * -cbf CBFORMAT --- format for CB-axis (e.g. %.2f)

    # * -u USING      --- Gnuplot's "using" splot specification
    # * -e EXTRA      --- extra Gnuplot commands/configuration, e.g., -e {set nokey; set title 'simple plot'}
    # * -w WITH       --- Gnuplot's "with" splot specification (used for contours)
    # * -t TERMINAL   --- terminal for the plot
    # * -v            --- visualize the generated image(s) (for file terminals)    
    #
    # ARGUMENT
    # * DATAFN --- datafile or function to plot
    #
    # PURPOSE
    # A fast way of making a heatmap from the x, y, f(x,y) tabulated
    # data. For a more elaborate and configurable 3D plotting, use the
    # :pwtk::gp::splot object-oriented interface instead.    
    #
    # If the datafile was created with the ::pwtk::write command, it
    # is closed prior to being plotted.
    #
    # The PWTK's predefined color palettes can be accessed with the
    # -pal option. To invert a palette, use "-pal {PALETTE negative}",
    # where PALETTE is one of:
    #
    # image:pwtk-palette.png
    # 
    # Note that for PALETTE only a substring that uniquely identifies it is
    # required.  For example, "-pal pa" selects the "pale-rainbow" palette.
    #
    # RETURN VALUE
    # * the name of created image (for file terminals)
    # * an empty string (for other terminals)
    #
    # SOURCE

    variable write_fid
    variable plot_id

    # parse args

    set narg 1
    ::pwtk::splotParseOpt_
    
    # datafile or function ?

    if { [file exists $args] } {
        # lastarg is a file
        finish $args
        set head   heatmap-[file rootname $args]
        set datafn [squote $args]
    } else {
        # lastarg is a function
        set id [incr ::pwtk::gp::heatmap_id]
        set head   heatmap-$id
        set datafn $args
    }

    # options post-processing

    set opts ""
    ifnotempty opt(pal) { append opts "[list -pal $opt(pal)] " }
    ifnotempty opt(cc)  { append opts "[list -cc $opt(cc)] " }
    if { ! $opt(contour) } { append opts "-nocontour " }
    
    ifnotempty opt(clw) { set opt(clw) "lw $opt(clw)" }
    ifnotempty opt(u)   { set opt(u)   "u $opt(u)" }
    ifset opt(w) {l nosurf}
    set with [concat w $opt(w) $opt(clw)]

    # create Gnuplot splot-map object
    
    set gp [::pwtk::gp::splot new -type map {*}$opts $head.$opt(t)]

    # add options + add excerpts

    ::pwtk::splotOptionsAdd_
    
    # plot
    
    if { $opt(contour) } {
        $gp plot [concat $datafn $opt(u), $datafn $opt(u) $with]
    } else {
        $gp plot [concat $datafn $opt(u)]
    }
    $gp exec    
    set output [$gp outputs]
    $gp destroy
    
    displayPlots_ $opt(v) $opt(t) $output
    
    return $output
}
#******



#****f* 5.8. Parameter Scans/::pwtk::splot
# SYNOPSIS
proc ::pwtk::splot {args} {    
    # USAGE
    #   ::pwtk::splot  ?OPTIONS?  DATAFN
    #
    # where DATAFN is either a datafile or a function to plot, and
    # OPTIONS are:
    #
    #    -type TYPE
    #    -pal  PALETTE
    #    -i    INTERPOLATION
    #    -contour
    #    -cc   CONTOUR_COLOR
    #    -clw  CONTOUR_LINEWIDTH
    #    -xl   XLABEL  
    #    -yl   YLABEL  
    #    -zl   ZLABEL  
    #    -xr   XRANGE  
    #    -yr   YRANGE  
    #    -zr   ZRANGE  
    #    -cbr  CBRANGE  
    #    -xf   XFORMAT 
    #    -yf   YFORMAT 
    #    -zf   ZFORMAT 
    #    -cbf  CBFORMAT 
    #    -u    USING    
    #    -w    WITH     
    #    -e    EXTRA    
    #    -t    TERMINAL 
    #    -v           
    #
    # OPTIONS
    # Options specific for 3D surface plots:
    # * -type TYPE    --- type of the Gnuplot splot: surface | both (default = surface; "TYPE = both" means surface + map)
    # * -pal PALETTE  --- color palette for the splot (PWTK has a dozen predefined palettes, see the image below)
    # * -i NX,NY      --- interpolation degree along X- and Y-axis
    #                     (default is set to auto-optimal interpolation)
    # * -contour      --- plot contours
    # * -cc  CONTOUR_COLOR --- assigns contour color
    # * -clw CONTOUR_LINEWIDTH --- contour linewidth
    #
    # General options:
    # * -xl XLABEL  --- X-axis label
    # * -yl YLABEL  --- Y-axis label
    # * -zl ZLABEL  --- Z-axis label

    # * -xr XRANGE    --- X-axis range specified as XMIN:XMAX or {XMIN:XMAX} or "XMIN:XMAX"
    # * -yr YRANGE    --- Y-axis range specified as YMIN:YMAX or {YMIN:YMAX} or "YMIN:YMAX"
    # * -zr ZRANGE    --- Z-axis range specified as ZMIN:ZMAX or {ZMIN:ZMAX} or "ZMIN:ZMAX"
    # * -cbr CBRANGE  --- CB-axis range specified as CBMIN:CBMAX or {CBMIN:CBMAX} or "CBMIN:CBMAX"
    
    # * -xf XFORMAT --- format for X-axis (e.g. %.2f)
    # * -yf YFORMAT --- format for Y-axis (e.g. %.2f)
    # * -zf ZFORMAT --- format for Z-axis (e.g. %.2f)
    # * -cbf CBFORMAT --- format for CB-axis (e.g. %.2f)
    
    # * -u USING    --- Gnuplot's "using" splot specification
    # * -w WITH     --- Gnuplot's "with" splot specification
    # * -e EXTRA    --- extra Gnuplot commands/configuration, e.g., -e {set nokey; set title 'simple plot'}
    # * -t TERMINAL --- terminal for the plot
    # * -v          --- visualize the generated image(s) (for file terminals)    
    #
    # ARGUMENT
    # * datafile --- datafile to plot
    #
    # PURPOSE
    # A fast way of making an splot from the x, y, f(x,y) tabulated
    # data. For a more elaborate 3D plotting, use the
    # ::pwtk::gp::splot object-oriented interface instead.
    #
    # If the datafile was created with the ::pwtk::write command, it
    # is closed prior to being plotted.
    #
    # The PWTK's predefined color palettes can be accessed with the
    # -pal option. To invert a palette, use "-pal {PALETTE negative}",
    # where PALETTE is one of:
    #
    # image:pwtk-palette.png
    #
    # Note that for PALETTE only a substring that uniquely identifies it is
    # required.  For example, "-pal pa" selects the "pale-rainbow" palette.
    #
    # RETURN VALUE
    # * the name of created image (for file terminals)
    # * an empty string (for non-file terminals)
    #
    # SOURCE

    variable write_fid
    variable plot_id

    # parse args

    set narg 1
    ::pwtk::splotParseOpt_
    
    # datafile or function ?

    if { [file exists $args] } {
        # lastarg is a file
        finish $args
        set head   splot-[file rootname $args]
        set datafn [squote $args]
    } else {
        # lastarg is a function
        set id [incr ::pwtk::gp::splot_id]
        set head   splot-$id
        set datafn $args
    }

    # options post-processing

    set opts ""
    ifnotempty opt(pal) { append opts "[list -pal $opt(pal)] " }
    ifnotempty opt(cc)  { append opts "[list -cc $opt(cc)] " }
    if { ! $opt(contour) } { append opts "-nocontour " }
    
    ifnotempty opt(clw) { set opt(clw) "lw $opt(clw)" }
    ifnotempty opt(u)   { set opt(u)   "u $opt(u)" }
    ifset opt(w) l 
    set with [concat w $opt(w) $opt(clw)]

    # create Gnuplot splot object

    set gp [::pwtk::gp::splot new -type $opt(type) {*}$opts $head.$opt(t)]

    # add options + add excerpts

    ::pwtk::splotOptionsAdd_

    # plot
    
    $gp plot [concat $datafn $opt(u) $with]
    $gp exec    
    set output [$gp outputs]
    $gp destroy
    
    displayPlots_ $opt(v) $opt(t) $output
    
    return $output
}
#******
