#
# $RCSfile: types.tcl,v $ --
#
#      This file contains definition of different numeric types (in
#      addition to Tcl's integer and double), such as posint, negint, ...
#
# Copyright (c) 2023  Anton Kokalj   Email: tone.kokalj@ijs.si
#
#
# This file is distributed under the terms of the GNU General Public
# License. See the file `LICENCE' in the root directory of the present
# distribution, or http://www.gnu.org/copyleft/gpl.txt .
#
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# ANTON KOKALJ BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
# AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
#
# $Id: header.tcl,v 1.3 2023/01/23 12:00:01 kokalj Exp $ 
#

# numeric types in addition to 'integer' and 'double':
#       binary
#       fboolean
#       boolean
#       int
#       posint      
#       nonposint   
#       negint      
#       nonnegint   
#       real | double | float
#       posreal | posdouble | posfloat
#       nonposreal | nonposdouble | nonposfloat  
#       negreal | negdouble | negfloat            
#       nonnegreal | nonnegdouble | nonnegfloat            


# internal...
proc ::pwtk::type::args_ {} {
    uplevel 1 {
        set narg [llength $args]
        if { $narg < 2 || $narg > 3 } {
            ::error "wrong # args: must be \"[::pwtk::procName] type ?-strict? number\", but got: \"[::pwtk::procName] $args\""
        }
        if { $narg == 3 } {
            if { [lindex $args 1] ne "-strict" } {
                ::error "wrong option [lindex $args 1], must be -strict"
            }
            set strict -strict
            set type   [lindex $args 0]
            set number [lindex $args 2]
        } else {
            set strict {}
            set type   [lindex $args 0]
            set number [lindex $args 1]
        }
    }
}


#****f* ::pwtk::type/::pwtk::type::number
# SYNOPSIS
proc ::pwtk::type::number {args} {
    # USAGE
    #   ::pwtk::type::number type ?-strict? value
    #
    # PURPOSE
    #   Query if a value (or number) is of corrected type. The
    #   following types are supported:
    #       integer | int
    #       double
    #       boolean
    #       fboolean
    #       binary
    #       posint      
    #       nonposint   
    #       negint      
    #       nonnegint   
    #       real | double | float
    #       posreal | posdouble | posfloat
    #       nonposreal | nonposdouble | nonposfloat  
    #       negreal | negdouble | negfloat            
    #       nonnegreal | nonnegdouble | nonnegfloat            
    # OPTION
    # * -strict -- if this option is specified, the empty value does not much
    #
    # ARGUMENTS
    # * type -- the type
    # * value -- value for which to test the type
    #
    # RETURN VALUE
    #   1 or 0
    #
    # SOURCE
    
    args_

    # perhaps number is specified in Fortran syntax, e.g., 1.0d0
    set number [::pwtk::f2c_number $number]
    
    if { $strict eq {} && $number eq {} } {
        # for non-strict, empty string matches all types
        return 1
    }
    
    switch -glob -- $type {
        fboolean {
            # boolean + Fortran logic (.t. .f. .true. .false.)
            return [string is boolean -strict [string trim $number .]]
        }
        boolean {
            return [string is boolean -strict $number]
        }
        *int* - binary {
            if { ! [string is integer -strict $number] } {
                return 0
            }
        }
        *real - *double - *float {
            if { ! [string is double -strict $number] } {
                return 0
            }
        }
    }
    switch -glob -- $type {
        binary {
            # 0 or 1
            if { $number == 0 || $number == 1 } {
                return 1
            } else {
                return 0
            }
        }
        nonpos* {
            return [expr $number <= 0]
        }
        nonneg* {
            return [expr $number >=0]
        }
        pos* {
            return [expr $number > 0]
        }
        neg* {
            return [expr $number < 0]
        }
        int* {
            return [string is integer -strict $number]
        }
        real - double - float {
            return [string is double -strict $number]
        }
    }
    
    # if we came here somting is wrong:
    ::error "unsupported type $type" 1
}
#******

#****f* ::pwtk::type/::pwtk::type::numberlist
# SYNOPSIS
proc ::pwtk::type::numberlist {args} {
    # USAGE
    #    ::pwtk::type::numberlist type ?-strict? numberlist
    # PURPOSE
    #   Query if the elements in the numberlist are of corrected
    #   type. For the supported types, see ::pwtk::type::number
    #
    # OPTION
    # * -strict -- if this option is specified, empty valuee do not much
    #
    # ARGUMENTS
    # * type -- the type
    # * numberlist -- list of numbers for which to test the type
    #
    # RETURN VALUE
    #   1 or 0
    #
    # SOURCE

    args_

    # list elements can be comma or space separated
    foreach n [regsub -all , $number { }] {
        if { ! [number $type {*}$strict $n] } {
            return 0
        }
    }
    return 1
}
#******


#****f* ::pwtk::type/::pwtk::type::optionlist
# SYNOPSIS
proc ::pwtk::type::optionlist {args} {
    # USAGE
    #   ::pwtk::type::optionlist allowedValues ?-strict? value
    #
    # PURPOSE
    #   Query if the value is one among the allowed values (allowed
    #   values are values of the optionlist).
    #
    # OPTION
    # * -strict -- if this option is specified, the empty value does not much
    #
    # ARGUMENTS
    # * allowedValues -- list of allowed values (aka optionlist)
    # * value -- value to check (it must be one among the allowed values)
    #
    # RETURN VALUE
    #   1 or 0
    #
    # SOURCE

    args_
    set allowedValues $type
    set value $number
    if { $strict eq {} && $value eq {} } {
        return 1
    } else {        
        return [expr {$value in $allowedValues}]
    }
}
#******


#****f* ::pwtk::type/::pwtk::type::range
# SYNOPSIS
proc ::pwtk::type::range {type from to number} {
    # PURPOSE
    #   Check if a 'number' is of correct type and is in the [from,to] range.
    # RETURN VALUE
    # * 1 if number is of correct type and in range
    # * 0 otherwise
    # SOURCE
    if { ! [::pwtk::type::number $type -strict $from] || ! [::pwtk::type::number $type -strict $to]
         ||  $number < $from || $number > $to } {
        return 0
    } else {
        return 1
    }
}
#******
